from datetime import date, timedelta
from .const import TODAY, TOMORROW
from statistics import mean, stdev


def get_offset_dict(offset_dict, dt_now) -> dict:
    return {
        TODAY: offset_dict.get(dt_now.date(), {}),
        TOMORROW: offset_dict.get(dt_now.date() + timedelta(days=1), {}),
    }


def set_offset_dict(prices: list[float], day: date) -> dict:
    ret = {}
    _len = len(prices)
    today: list = []
    tomorrow: list = []
    match _len:
        case 23 | 24 | 25 | 92 | 96 | 100:
            today = prices
        case 47 | 48 | 49 | 188 | 192 | 196:
            today = prices[: len(prices) // 2]
            tomorrow = prices[len(prices) // 2 : :]
    ret[day] = _deviation_from_mean(today, prices)
    ret[day + timedelta(days=1)] = _deviation_from_mean(tomorrow, prices)
    return ret


def _deviation_from_mean(prices: list[float], checker: list[float]) -> dict[int, float]:
    if not len(prices):
        return {}
    avg = mean(checker)
    devi = stdev(prices)
    deviation_dict = {}
    for i, num in enumerate(prices):
        deviation = (num - avg) / devi
        if devi < 1:
            deviation *= 0.5
        deviation_dict[i] = round(deviation, 2)
    return deviation_dict



from .savings_model import SavingsModel
from .savings_status import SavingsStatus
from typing import Tuple
from datetime import date, datetime, timedelta
import logging

_LOGGER = logging.getLogger(__name__)


class SavingsService:
    def __init__(self, peak_price: float) -> None:
        self.model = SavingsModel(peak_price_per_kwh=peak_price)
        self._savings_peak = 0
        self._savings_trade = 0

    @property
    def status(self) -> SavingsStatus:
        return self.model.status

    @property
    def savings_peak(self) -> float:
        return self._savings_peak

    @savings_peak.setter
    def savings_peak(self, value: float) -> None:
        self._savings_peak = round(value, 2)

    @property
    def savings_trade(self) -> float:
        return self._savings_trade

    @savings_trade.setter
    def savings_trade(self, value: float) -> None:
        self._savings_trade = round(value, 2)

    @property
    def savings_total(self) -> float:
        return sum([self.savings_peak, self.savings_trade])

    async def async_export_data(self) -> dict:
        return {
            "car_connected_at": self.model.car_connected_at,
            "prices": self.model.prices,
            "consumption": self.model.consumption,
            # "peaks": self.model.peaks,
        }

    async def async_import_data(self, data: dict) -> None:
        try:
            self.model.car_connected_at = data["car_connected_at"]
            self.model.prices = data["prices"]
            self.model.consumption = data["consumption"]
        except Exception as e:
            _LOGGER.error(f"Could not import data: {e}")

    async def async_start_listen(self, connected_at: datetime | None = None) -> None:
        self.model.car_connected_at = connected_at or datetime.now()
        self.model.status = SavingsStatus.Collecting

    async def async_stop_listen(self) -> None:
        await self.model.async_reset()

    async def async_add_prices(
        self, prices: list[float], _date: date | None = None
    ) -> None:
        await self.model.async_add_prices(prices, _date)

    async def async_add_to_consumption(
        self,
        consumption: float,
        _date: date | None = None,
        _hour: int | None = None,
    ) -> None:
        await self.model.async_add_to_consumption(consumption, _date, _hour)

    async def async_register_charge_session(
        self, charge_session: dict, original_peak: float | None = None
    ) -> None:
        estimated_power, total_energy = await self.async_calculate_estimated_power(
            charge_session
        )
        simulated_cost = await self.async_direct_charge_cost(
            estimated_power, total_energy
        )
        charge_cost = await self.async_calculate_charge_cost(charge_session)
        peaks_increase = await self.async_direct_peak_cost(
            charge_session, original_peak
        )
        self.savings_peak = peaks_increase
        self.savings_trade = max(simulated_cost - charge_cost, 0)

    async def async_direct_peak_cost(
        self, charge_session: dict, original_peak: float | None
    ) -> float:
        ret = 0
        if original_peak is None:
            return ret
        observed_peak = original_peak
        for key, value in charge_session.items():
            peaks = self.model.consumption.get(key, None)
            if peaks is None:
                _LOGGER.error(f"Could not find peaks for date {key}.")
                return 0
                # raise Exception
            for hour, energy in value.items():
                if energy + self.model.consumption[key][hour] > observed_peak:
                    ret = energy + self.model.consumption[key][hour]
                    observed_peak = energy + self.model.consumption[key][hour]
        diff = ret - original_peak
        if diff < 0:
            return 0
        return diff * self.model.peak_price_per_kwh

    async def async_direct_charge_cost(
        self, estimated_power: int, total_energy: float
    ) -> float:
        ret = 0
        dt = self.model.car_connected_at
        if dt is None:
            dt = datetime.now()
        while total_energy > 0:
            power_key = estimated_power / 1000
            ret += min(total_energy, power_key) * self.model.prices[dt.date()][dt.hour]
            total_energy -= min(total_energy, power_key)
            dt += timedelta(hours=1)
        return ret

    async def async_calculate_charge_cost(self, charge_session: dict) -> float:
        ret = 0
        for key, value in charge_session.items():
            prices = self.model.prices.get(key, None)
            if prices is None:
                raise Exception
            for hour, power in value.items():
                ret += power * prices[hour]
        return ret

    async def async_calculate_estimated_power(
        self, charge_session: dict
    ) -> Tuple[int, float]:
        ret = {}
        ret_idx = 0
        for key, value in charge_session.items():
            for hour, power in value.items():
                ret[ret_idx] = power * 1000
                ret_idx += 1
        try:
            return int(sum(ret.values()) / len(ret)), sum(ret.values()) / 1000
        except ZeroDivisionError:
            """there has been no charging. return 0. Should this be decided before this function?"""
            # _LOGGER.error("Could not calculate estimated power for savings.")
            return 0, 0



from __future__ import annotations
import time
from datetime import datetime, date
from .power_reading import PowerReading
from .energy_weekly import EnergyWeekly
from .session_model import SessionModel
from typing import Any
import logging

_LOGGER = logging.getLogger(__name__)


class SessionService:
    def __init__(self) -> None:
        self.model = SessionModel()
        self.average_data = EnergyWeekly()
        self._original_peak: float = 0
        self._current_date: date = date.today()
        self._current_hour: int = 0

    async def async_setup(
        self, init_average_data: dict | None = None, mock_time: float | None = None
    ) -> None:
        await self.average_data.async_setup(init_average_data)
        await self.async_set_delta(mock_time)

    @property
    def energy_average(self) -> float:
        return self.average_data.average

    @property
    def energy_weekly_dict(self) -> dict:
        return self.average_data.export

    @property
    def total_energy(self) -> float:
        return round(self.sync_get_status()["energy"]["value"], 3)

    @property
    def readings(self) -> list:
        return self.model.readings

    @property
    def total_price(self) -> float:
        return round(self.sync_get_status()["price"], 3)

    @total_price.setter
    def total_price(self, val):
        self.total_price = val

    @property
    def session_data(self) -> dict:
        return self.model.consumption

    @property
    def original_peak(self) -> float:
        return self._original_peak

    async def async_reset(self, original_peak: float = 0):
        self._original_peak = original_peak
        _init_avg = self.average_data.export
        self.__init__()
        await self.async_setup(_init_avg)

    async def async_terminate(self, mock_time: float | None = None):
        _now = mock_time or time.time()
        await self.async_update_power_reading(0, _now)
        _LOGGER.debug(
            f"Called terminate on session_price. Trying to add {self.total_energy} to statistics."
        )
        await self.average_data.async_update(self.total_energy, _now)
        await self.async_get_status()

    async def async_set_status(self) -> None:
        for i in self.model.readings:
            self.model.total_energy += i.reading_integral
            self.model.total_price += i.reading_cost
        await self.model.async_add_consumption(
            consumption=self.model.total_energy,
            date=self._current_date,
            hour=self._current_hour,
        )

    async def async_get_status(self) -> dict:
        self.model.total_energy = 0
        self.model.total_price = 0
        await self.async_set_status()
        return {
            "energy": {"value": self.model.total_energy, "unit": "kWh"},
            "price": self.model.total_price,
        }

    def sync_set_status(self) -> None:
        for i in self.model.readings:
            self.model.total_energy += i.reading_integral
            self.model.total_price += i.reading_cost

    def sync_get_status(self) -> dict:
        self.model.total_energy = 0
        self.model.total_price = 0
        self.sync_set_status()
        return {
            "energy": {"value": self.model.total_energy, "unit": "kWh"},
            "price": self.model.total_price,
        }

    async def async_update_power_reading(
        self, power: Any, mock_time: float | None = None
    ):
        await self.async_set_delta(mock_time or time.time())
        p = PowerReading(
            self.model.price, self.model.current_power, self.model.time_delta
        )
        self.model.readings.append(p)
        try:
            self.model.current_power = float(power)
        except:
            self.model.current_power = 0

    async def async_update_price(self, price: Any, mock_time: float | None = None):
        if self.model.current_power > 0:
            await self.async_update_power_reading(
                power=self.model.current_power, mock_time=mock_time or time.time()
            )
        try:
            self.model.price = float(price)
        except:
            self.model.price = 0

    async def async_set_delta(self, mock_time: float | None = None) -> None:
        now = mock_time or time.time()
        await self.async_update_date_and_hour(now)
        self.model.time_delta = now - self.model.current_time
        self.model.current_time = now

    async def async_update_date_and_hour(self, dt: float) -> None:
        date_time = datetime.fromtimestamp(dt)
        self._current_date = date_time.date()
        self._current_hour = date_time.hour



from dataclasses import dataclass, field
from .hour_price import HourPrice
from datetime import datetime
from statistics import mean

@dataclass
class MaxMinModel:
    min_price: float
    input_hours: list[HourPrice] = field(default_factory=lambda: [])
    expected_hourly_charge: float = 0

    def caluclate_average_price(
        self,
        input: list[HourPrice],
        total_charge: float | None,
        peak: float | None,
    ) -> float | None:
        #_total = total_charge or 0
        #_peak = peak or self.expected_hourly_charge
        try:
            first = mean(
                [
                    hp.price
                    for hp in input
                    if hp.permittance > 0 and not hp.passed
                ]
            )
            return round(first, 2)
        except ZeroDivisionError as e:
            return 0.0

    def calculate_total_charge(self, input: list[HourPrice]) -> float:
        return round(
            sum(
                [
                    hp.permittance * self.expected_hourly_charge
                    for hp in input
                    if not hp.passed
                ]
            ),
            1,
        )



"""Peak querytypes"""
QUERYTYPE_BASICMAX = "BasicMax"
QUERYTYPE_AVERAGEOFTHREEDAYS = "AverageOfThreeDays"
QUERYTYPE_AVERAGEOFTHREEHOURS = "AverageOfThreeHours"
QUERYTYPE_AVERAGEOFTHREEDAYS_MIN = "AverageOfThreeDays_Min"
QUERYTYPE_AVERAGEOFTHREEHOURS_MIN = "AverageOfThreeHours_Min"
QUERYTYPE_AVERAGEOFFIVEDAYS = "AverageOfFiveDays"
QUERYTYPE_AVERAGEOFFIVEDAYS_MIN = "AverageOfFiveDays_Min"
QUERYTYPE_HIGHLOAD = "HighLoad"
QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19 = "sala"
QUERYTYPE_AVERAGEOFTHREEHOURS_MON_FRI_07_19_MIN = "sala"
QUERYTYPE_MAX_NOV_MAR_MON_FRI_06_22 = "skövde"
QUERYTYPE_BASICMAX_MON_FRI_07_17_DEC_MAR_ELSE_REGULAR = "kristinehamn"
QUERYTYPE_SOLLENTUNA = "sollentuna"
QUERYTYPE_SOLLENTUNA_MIN = "sollentuna_min"
QUERYTYPE_NEVER = "nopeak"



from __future__ import annotations
from typing import TYPE_CHECKING, Callable

if TYPE_CHECKING:
    from ....models.hourselection.hourselection_options import HourSelectionOptions
from dataclasses import dataclass, field
from .hour_price import HourPrice
from datetime import date, datetime, time, timedelta
from .list_type import ListType
from ...hourselection.hourselectionservice.hoursselection_helpers import convert_none_list
from ..offset_dict import get_offset_dict, set_offset_dict


@dataclass
class HourSelectionServiceModel:
    _prices_today: list[float] = field(default_factory=list)
    _prices_tomorrow: list[float] = field(default_factory=list)
    _hours_prices: list[HourPrice] = field(default_factory=list)
    offset_dict: dict[datetime, dict] = field(default_factory=dict)
    adjusted_average: float | None = None
    use_quarters: bool = False

    @property
    def hours_prices(self) -> list[HourPrice]:
        return self._hours_prices

    @hours_prices.setter
    def hours_prices(self, val: list[HourPrice]) -> None:
        if val is not None:
            if len(val) > 0:
                self._hours_prices = val
                self._hours_prices.sort(key=lambda x: x.dt)
        else:
            self._hours_prices = []

    @property
    def prices_today(self) -> list:
        return self._prices_today

    @prices_today.setter
    def prices_today(self, val):
        self._prices_today = val
        if self._prices_today == self.prices_tomorrow:
            self.prices_tomorrow = []

    @property
    def prices_tomorrow(self) -> list:
        return self._prices_tomorrow

    @prices_tomorrow.setter
    def prices_tomorrow(self, val):
        ret = convert_none_list(val)
        self._prices_tomorrow = ret

    def set_hourprice_lists(self, prices:list, prices_tomorrow:list, service_options:HourSelectionOptions, datum:date, datum_tomorrow:date, is_passed_func:Callable) -> None:
        self.set_hourprice_list(prices, service_options, datum, is_passed_func)
        self.set_hourprice_list(prices_tomorrow, service_options, datum_tomorrow, is_passed_func)
        self.hours_prices.sort(key=lambda x:x.dt)

    def set_hourprice_list(
        self,
        prices: list,
        service_options: HourSelectionOptions,
        datum: date,
        is_passed_func: Callable,
    ) -> None:
        ret = []
        for idx, p in enumerate(prices):  # type: ignore
            assert isinstance(p, (float, int))
            hour = int(idx / 4) if self.use_quarters else idx
            quarter = idx % 4 if self.use_quarters else 0
            rret = HourPrice(
                dt=self._get_dt(datum, hour, quarter),
                quarter=quarter,
                price=p,
                passed=is_passed_func(datum, hour, quarter),
                hour_type=HourPrice.set_hour_type(
                    service_options.absolute_top_price, service_options.min_price, p
                ),
                list_type=ListType.Quarterly if self.use_quarters else ListType.Hourly,
            )
            if rret not in self.hours_prices:
                ret.append(rret)
        if len(self.hours_prices) > 0:
            self.hours_prices.extend(ret)
        else:
            self.hours_prices = ret

    def get_offset_dict(self, dt_date: datetime) -> dict:
        return get_offset_dict(self.offset_dict, dt_date)

    def set_offset_dict(self, prices: list[float], day: date) -> None:
        if all(
            [
                day in self.offset_dict.keys(),
                day - timedelta(days=1) in self.offset_dict.keys(),
            ]
        ):
            return
        self.offset_dict = set_offset_dict(prices, day)

    def get_future_hours(self, dtmodel) -> list[HourPrice]:
        for hp in self.hours_prices:
            hp.set_passed(dtmodel)
        return [hp for hp in self.hours_prices if not hp.passed]

    @staticmethod
    def _get_dt(datum: date, hour: int, quarter: int) -> datetime:
        return datetime.combine(
            datum, time(hour=hour, minute=quarter * 15, second=0, microsecond=0)
        )


from __future__ import annotations
from typing import TYPE_CHECKING

from ...models.hourselection.cautionhourtype import CautionHourType

import logging
from dataclasses import dataclass, field
from .topprice_type import TopPriceType
from .set_top_price import (
    set_absolute_top_price,
    async_add_tomorrow,
    async_validate_top_min_prices,
)

_LOGGER = logging.getLogger(__name__)

from datetime import datetime
from statistics import mean
from calendar import monthrange


@dataclass(frozen=False)
class HourSelectionOptions:
    cautionhour_type_enum: CautionHourType = CautionHourType.SUAVE
    top_price: float = 0  # move to separate file
    fixed_top_price: float = 0  # move to separate file
    dynamic_top_price: float = 0  # move to separate file
    top_price_type: TopPriceType = field(default_factory=lambda: TopPriceType.Unset)
    min_price: float = 0
    absolute_top_price: float = field(init=False)  # move to separate file

    def __post_init__(self):
        self.fixed_top_price = self.top_price
        (
            self.absolute_top_price,
            self.top_price_type,
            self.min_price,
        ) = set_absolute_top_price(self.min_price, self.top_price)

    async def async_add_tomorrow_to_top_price(
        self, prices_tomorrow: list, mock_day: int | None = None
    ) -> float:  # move to separate file
        _day = mock_day or datetime.now().day
        if _day + 1 > monthrange(datetime.now().year, datetime.now().month)[1]:
            """tomorrow is a new month so those prices should be calculated separately"""
            return mean(prices_tomorrow)
        if self.top_price_type != TopPriceType.Dynamic:
            return self.absolute_top_price
        return min(
            await async_add_tomorrow(_day, prices_tomorrow, self.absolute_top_price),
            self.absolute_top_price,
        )

    async def async_set_absolute_top_price(
        self, monthly_avg_top: float | None = None
    ) -> None:
        _min = self.min_price
        self.fixed_top_price = self.top_price
        if monthly_avg_top is None:
            top = self.top_price
        else:
            self.dynamic_top_price = monthly_avg_top
            top = min(monthly_avg_top, self.top_price)
            if monthly_avg_top <= self.top_price:
                self.top_price_type = TopPriceType.Dynamic
            else:
                self.top_price_type = TopPriceType.Absolute

        if not await async_validate_top_min_prices(top, _min):
            _LOGGER.warning(
                f"Setting top-price and min-price to zero because of min-price being larger than top-price. Please fix in options. top:{top} min:{_min}"
            )
            top = 0
            HourSelectionOptions.min_price = 0
        if top is None:
            self.absolute_top_price = float("inf")
        elif top <= 0:
            self.absolute_top_price = float("inf")
        else:
            self.absolute_top_price = float(top)



from ....services.locale.querytypes.queryservice import QueryService
from ....models.locale.enums.querytype import QueryType
from ..time_pattern import TimePattern
from ....models.locale.enums.calendar_periods import CalendarPeriods
from ..locale_model import Locale_Type
from dataclasses import dataclass
from ..querytypes.querysets import QUERYSETS
from ..querytypes.querytypes import QUERYTYPES
from ..querytypes.const import QUERYTYPE_NEVER


@dataclass
class Default(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]


@dataclass
class NoPeak(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.Max
        self.charged_peak = QueryType.Max
        self.query_model = QUERYTYPES[QueryType.Max]
        self.query_service = QueryService(QUERYSETS[QUERYTYPE_NEVER])
        self.free_charge_pattern = TimePattern(
            [
                {
                    CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                    CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                    CalendarPeriods.Hour: [
                        0,
                        1,
                        2,
                        3,
                        4,
                        5,
                        6,
                        7,
                        8,
                        9,
                        10,
                        11,
                        12,
                        13,
                        14,
                        15,
                        16,
                        17,
                        18,
                        19,
                        20,
                        21,
                        22,
                        23,
                    ],
                }
            ]
        )
